#################################################################################
# Created       : Martin Foster
# Modified      : 28-Nov-2005
#################################################################################
#
# Image - Image handling module for Ethereal Realms 
# Copyright (C) 2000-2006 Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

package Ethereal::Image;				# Package name

use Image::Magick;					# Image manipulation
use Image::ExifTool;					# Image EXIF information 


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;				# Class
        my $database = shift;				# Database handle
	my $basedir  = shift;				# Base directory

	my $self;					# Reference to self

	my %system;					# System hash


	#####################
	# Program files

        # Link with hash
        $database->GetHashSystem(\%system);


	# Base directory
	$basedir = (defined($basedir)) ? $basedir : $system{'SetBaseDir'};


	# Generate hash
        %{$self}  = (
	       HANDLE  => $database,			# Database handle
               BASEDIR => $basedir,			# Default directory
	       PIXEL   => $system{'SetThumbnail'},      # Thumbnail demensions
	       QUALITY => $system{'SetImgQuality'},	# Quality of image
	       BGROUND => $system{'SetImgBackground'},	# Background colour of image
	       FORMAT  => $system{'SetImgType'},	# Image type
	       VIEWMAX => $system{'SetImgView'},	# Image view size
	       FULLMAX => $system{'SetImgFull'},	# Image view full image size
	       SEP     => '/'                           # For SEP use '/' for Unix and '\' for Windows
	 );


	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}


####################
# Exif
#
# Wrapper designed to hide the complexity of the EXIF tag
# extraction taking place when an image is uploaded.

sub Exif
{
	####################
	# Data members

	my $self    = shift;				# Self referencing link
	my $image   = shift;				# Specific image

	my $basedir = $self->{'BASEDIR'};		# Base working directory
	my $sep     = $self->{'SEP'};			# Seperator

	my $exiftool;					# Exif tag handle
	my $info;					# Info file

	####################
	# Program area

	# Create instace
	$exiftool = new Image::ExifTool;

	# Set options
	$exiftool->Options(
		Compact    => 1,
		DateFormat => '%d-%b-%g %H:%M:%S%Z',
		Duplicates => 0,
		Sort       => 'File'
	 );

	# Extract information
	$info = $exiftool->ImageInfo("$basedir$sep$image");

	# Return handle
	return $info;
}


####################
# Exif Format
#
# Wrapper designed to hide the complexity of the EXIF tag
# extraction taking place when an image is uploaded. This
# will additionally take a file format it into a singular
# line using CSS for compatibility.

sub ExifFormat
{
	####################
	# Data members

	my $self     = shift;				# Self referencing link
	my $image    = shift;				# Specific image

	my $database = $self->{'HANDLE'};		# Database handle				

	my $statement;					# Database statement
	my $res;					# Database results
	
	my $exif;					# Exif handle
	my $out;					# Output template


	####################
	# Program area

	# Extract information
	$exif = $self->Exif($image);	

	# Safety precaution
	if (defined($exif))
	{
		# Initialize template
		($out) = $database->DocumentGetGalleryExif();

		# Fetch appropriatly
		$database->Pull(\$statement, "SELECT
			 ExifTag AS \"Tag\",
			 ExifMacro AS \"Macro\"
			FROM Exif
			ORDER BY ExifTag");

		# Loop and fetch
		while ($res = $statement->fetchrow_hashref())
		{
			# Normal operation
			if ((defined($exif->{$res->{'Tag'}}))
			 && (length($exif->{$res->{'Tag'}}) > 0))
			{
				# Replace in template
				$out =~ s/$res->{'Macro'}/$exif->{$res->{'Tag'}}/gs;
			}

			# Gracefully leave
			else
			{
				# Undefine
				$out = undef;

				# Leave loop
				last;
			}
		}

		# Finish query
		$statement->finish();
		
		# Return
		return $out;
	}

	# Return null
	return undef;
}


####################
# Set
#
# Sets the size of the image to specific demensions as determined
# by the parameters received.

sub Set
{
	####################
	# Data members

	my $self    = shift;				# Self referencing link
	my $image   = shift;				# Specific image
	my $max     = shift;				# Max demensions			
	my $magick  = shift;				# Image handle
	my $width   = shift;				# Image width
	my $height  = shift;				# Image height

	my $basedir = $self->{'BASEDIR'};		# Base working directory
	my $pixel   = $self->{'PIXEL'};			# Pixel demensions
	my $sep     = $self->{'SEP'};			# Seperator



	####################
	# Program area

	# Create and initialize 
	# instances only if necessary
	unless (defined($magick))
	{
		# Create instance
		$magick = new Image::Magick;

		# Open
		$magick->Read("$basedir$sep$image");
	}

	
	# Pull dimensions
	($width, $height) = $self->Size($image, $magick) unless (defined($width));

	# Useable
	if ((defined($width))
	 && (defined($height)))
	{
		# Resize calculations
		# Height
		if ($height > $width)
		{
			# Calculate demensions
			$width  = int(($width * $max) / $height);
			$height = $max;
		}

		# Width
		else
		{
			# Calculate demensions
			$height = int(($height * $max) / $width);
			$width  = $max;
		}


		# Image manipulation
		# Resize operation
		$magick->Resize(width=>$width, height=>$height);

		# Write
		$self->Write("$image", $magick) if ($image ne 'bypass');
	}
}

# Alias for internal bypass
sub SetBypass
{ Set(shift, 'bypass', shift, shift, shift, shift); }


####################
# Size
#
# A simple file size handling subroutine that is used to gain various
# information on images.  Needed for basic chatting, thumbnail creation
# and image resizing.
#
# Due to possible problems related to the use of the Image::Magick
# library, there is an option for legacy support which will allow one
# to use Image::Size instead.

sub Size
{
	####################
	# Data members

	my $self    = shift;				# Self referencing link
	my $image   = shift;				# Specific image
	my $magick  = shift;				# Image handle

	my $basedir = $self->{'BASEDIR'};		# Base working directory
	my $sep     = $self->{'SEP'};			# Seperator

	my $format;					# Image format
	my $height;					# Image height
	my $size;					# Image size
	my $width;					# Image width


	####################
	# Program area

	# Create instance
	$magick = new Image::Magick unless (defined($magick));

	# Retreive parameters
	($width, $height, $size, $format) = $magick->Ping("$basedir$sep$image");

	# Work around old versions of Image::Magick
	($width, $height, $size, $format) = split(/,/, $width) if ($width =~ /\,/g);

	# File size in Kilobytes
	$size = int($size/1024);

	# Return values
	return $width, $height, $size, $format;
}


####################
# Thumbnail gen
#
# Used to generate thumbnails of images received.  This is accomplished by
# generating three seperate images, the user selects the best likeness
# using alternate means.

sub Thumbnail
{
	####################
	# Data members

	my $self    = shift;				# Self referencing link
	my $image   = shift;				# Specific image
	my $magick  = shift;				# Image handle

	my $basedir = $self->{'BASEDIR'};		# Base working directory
	my $pixel   = $self->{'PIXEL'};			# Pixel demensions
	my $sep     = $self->{'SEP'};			# Seperator

	my $format;					# Image format
	my $height;					# Image height
	my $size;					# Image size in Kilobytes
	my $width;					# Image width

	my $owidth;					# Original width
	my $oheight;					# Original width
	
	my $ibg;					# Image preset background colours
	my $iformat;					# Image formatting
	my $imax;					# Max file size
	my $iview;					# Viewing image

	my $x;						# X coordinates
	my $y;						# Y coordinates
	my $depth;					# Depth of an image
	my $ext;					# Extension

	my $mid;					# Mid way cut
	my $neg;					# Negative cut
	my $pos;					# Positive cut

	my $ful;					# Complete image
	my $tmp;					# Temporary work image

	my $return = 0;					# Return value


	####################
	# Program area

	# Create instance
	$magick = new Image::Magick;

	
	# Pull dimensions
	($width, $height, $size, $format) = $self->Size($image, $magick);


	# Open file for control
	$magick->Read("$basedir$sep$image");


	# Set
	$owidth  = $width;
	$oheight = $height;


	# Safety precautions
	$ibg     = (defined($self->{'BGROUND'})) ? $self->{'BGROUND'} : 'transparent';
	$iformat = (defined($self->{'FORMAT'}))  ? $self->{'FORMAT'}  : 'png';
	$imax    = (defined($self->{'FULLMAX'})) ? $self->{'FULLMAX'} : 800;
	$iview   = (defined($self->{'VIEWMAX'})) ? $self->{'VIEWMAX'} : 640;

	# Filename adapt
	$image =~ /(\w+)\.(\w+)$/;
	$image = $1;
	$ext   = $2;
	

	# Setting of depth
	$depth = ($iformat =~ /^(png|gif)$/i) ? 8 : 16;

	# Full demension image
	# Image handling for full sized thumbnail
	if ($ibg =~ /^(\#|\w)\w+$/)
	{
		# Create blank image with transparencies at full demension
		$ful = new Image::Magick;
		$ful->Set(size=>"${pixel}x${pixel}");
		$ful->Read("XC:$ibg");
		$ful->Set(magick=>uc($iformat));
	}
	else
	{
		# Simply load the background image
		$ful = Image::Magick->new();
		$ful->Read($ibg);
	}

	# Copy current image to temporary
	$tmp = $magick->Clone();

	# Resize copy accordingly
	$self->SetBypass($pixel, $tmp, $width, $height);
	
	# Composite image
	$ful->Composite(
		image     => $tmp, 
		compose   => 'over', 
		gravity   => 'center'
	 );

	# Write object
	$self->Write("thumb-$image-ful.$iformat", $ful, $depth);

	# Clear image
	undef $ful;
	undef $tmp;


	# Regular image resize
	# Set regular view accordingly
	if (($owidth > $iview)
	 || ($oheight > $iview))
	{
		# Clone
		$tmp = $magick->Clone();

		# Resize copy accordingly
		$self->SetBypass($iview, $tmp, $owidth, $oheight);

		# Write
		$self->Write("view-$image.$ext", $tmp);

		# Clear image
		undef $tmp;

		# Increment return
		$return++;
	}

	# Copy of regular instead
	else
	{
		# Write original file
		$self->Write("view-$image.$ext", $magick);
	}

	# Set full view accordingly
	if (($owidth > $imax)
	 || ($oheight > $imax))
	{
		# Clone
		$tmp = $magick->Clone();

		# Resize copy accordingly
		$self->SetBypass($imax, $tmp, $owidth, $height);

		# Write
		$self->Write("$image.$ext", $tmp);

		# Clear image
		undef $tmp;

		# Increment return
		$return++;
	}	
	
	# Resize calculations for square images
	# Height
	if ($height < $width)
	{
		# Calculate demensions
		$width  = int(($width * $pixel) / $height);
		$height = $pixel;

		# Crop co-ords
		$x = int(($width - $pixel) / 2);
		$y = 0;
	}

	# Width
	else
	{
		# Calculate demensions
		$height = int(($height * $pixel) / $width);
		$width  = $pixel;

		# Crop co-ords
		$x = 0;
		$y = int(($height - $pixel) / 2);
	}
	

	# Resize operation
	$magick->Resize(width=>$width, height=>$height);
	


	# Crop and deliver
	#
	# Negative
	# Work off a clone
	$tmp = $magick->Clone();
	$tmp->Crop(width=>$pixel, height=>$pixel, x=>0, y=>0);


	# Copy in a blank slate
	$neg = new Image::Magick;
	$neg->Set(size=>"${pixel}x${pixel}");
	$neg->Read("XC:$ibg");
	$neg->Composite($tmp);
	
	# Set format
	$neg->Set(magick=>uc($iformat));

	# Write
	$self->Write("thumb-$image-neg.$iformat", $neg, $depth);

	
	# Clear image
	undef $neg;
	undef $tmp;


	# Middle
	# Work off a clone
	$tmp = $magick->Clone();
	$tmp->Crop(width=>$pixel, height=>$pixel, x=>$x, y=>$y);

	# Copy into a blank slate
	$mid = new Image::Magick;
	$mid->Set(size=>"${pixel}x${pixel}");
	$mid->Read("XC:$ibg");
	$mid->Composite($tmp);

	# Set format
	$mid->Set(magick=>uc($iformat));

	# Write
	$self->Write("thumb-$image-mid.$iformat", $mid, $depth);

	# Clear image
	undef $mid;
	undef $tmp;


	# Double
	$x += $x;
	$y += $y;

	# Positive
	# Work off a clone
	$tmp = $magick->Clone();
	$tmp->Crop(width=>$pixel, height=>$pixel, x=>$x, y=>$y);

	# Cope into a blank slate
	$pos = new Image::Magick;
	$pos->Set(size=>"${pixel}x${pixel}");
	$pos->Read("XC:$ibg");
	$pos->Composite($tmp);

	# Set format
	$pos->Set(magick=>uc($iformat));

	# Write
	$self->Write("thumb-$image-pos.$iformat", $pos, $depth);

	# Clear image
	undef $pos;
	undef $tmp;
	

	# Clear image itself
	@$magick = ();
	


	# Return
	return $return;
}


####################
# Write
#
# Image writer, used to write the final changes of images when
# the time comes.

sub Write
{
	####################
	# Data members

	my $self    = shift;				# Self referencing link
	my $image   = shift;				# Specific image
	my $magick  = shift;				# Image handle
	my $depth   = shift;				# Image depth 

	my $basedir = $self->{'BASEDIR'};		# Base working directory
	my $quality = $self->{'QUALITY'};		# Quality of image
	my $sep     = $self->{'SEP'};			# Seperator


	####################
	# Program Area

	# Safety
	$quality = 30 unless (defined($quality));

	# True image name
	$image = $basedir . $sep . $image;

 	# Rewrite thumbnail
	eval
	{
		# Setting changes
		$magick->Set(quality  => $quality);
		$magick->Set(depth    => $depth) if (defined($depth));

		# Write image
		$magick->Write(filename => $image);

		# Return
		return 0;
	};

	# Error occured
	if ($@)
	{
		# Issue warning
		warn("Saving of image $image failed: $@");

		# Return error
		return $!;
	}
}

1;								# Return true
